@@ -1,22 +1,27 @@
 """This module contains the application class."""
 
 import json
+import random
 import shutil
 import time
 from dataclasses import dataclass
 from datetime import datetime
-from typing import Any, Dict, List, Optional, Union
+from typing import Any, Optional, Union
 
 import requests
 
 from nextmv.base_model import BaseModel
 from nextmv.cloud import package
-from nextmv.cloud.acceptance_test import AcceptanceTest, Metric
+from nextmv.cloud.acceptance_test import AcceptanceTest, ExperimentStatus, Metric
 from nextmv.cloud.batch_experiment import BatchExperiment, BatchExperimentMetadata, BatchExperimentRun
 from nextmv.cloud.client import Client, get_size
 from nextmv.cloud.input_set import InputSet
+from nextmv.cloud.instance import Instance, InstanceConfiguration
 from nextmv.cloud.manifest import Manifest
-from nextmv.cloud.status import Status, StatusV2
+from nextmv.cloud.run import ExternalRunResult, RunConfiguration, RunInformation, RunLog, RunResult
+from nextmv.cloud.secrets import Secret, SecretsCollection, SecretsCollectionSummary
+from nextmv.cloud.status import StatusV2
+from nextmv.cloud.version import Version
 from nextmv.logger import log
 from nextmv.model import Model, ModelConfiguration
 
@@ -32,98 +37,63 @@ class DownloadURL(BaseModel):
     """URL to use for downloading the file."""
 
 
-class ErrorLog(BaseModel):
-    """Error log of a run, when it was not successful."""
-
-    error: Optional[str] = None
-    """Error message."""
-    stdout: Optional[str] = None
-    """Standard output."""
-    stderr: Optional[str] = None
-    """Standard error."""
-
-
-class Metadata(BaseModel):
-    """Metadata of a run, whether it was successful or not."""
-
-    application_id: str
-    """ID of the application where the run was submitted to."""
-    application_instance_id: str
-    """ID of the instance where the run was submitted to."""
-    application_version_id: str
-    """ID of the version of the application where the run was submitted to."""
-    created_at: datetime
-    """Date and time when the run was created."""
-    duration: float
-    """Duration of the run in milliseconds."""
-    error: str
-    """Error message if the run failed."""
-    input_size: float
-    """Size of the input in bytes."""
-    output_size: float
-    """Size of the output in bytes."""
-    status: Status
-    """Deprecated: use status_v2."""
-    status_v2: StatusV2
-    """Status of the run."""
-
-
 class PollingOptions(BaseModel):
-    """Options to use when polling for a run result."""
+    """
+    Options to use when polling for a run result.
 
-    backoff: float = 1
-    """Backoff factor to use between polls. Leave this at 1 to poll at a
-    constant rate."""
-    delay: float = 1
-    """Delay to use between polls, in seconds."""
+    The Cloud API will be polled for the result. The polling stops if:
+
+    * The maximum number of polls (tries) are exhausted. This is specified by
+      the `max_tries` parameter.
+    * The maximum duration of the polling strategy is reached. This is
+      specified by the `max_duration` parameter.
+
+    Before conducting the first poll, the `initial_delay` is used to sleep.
+    After each poll, a sleep duration is calculated using the following
+    strategy, based on exponential backoff with jitter:
+
+    ```
+    sleep_duration = min(`max_delay`, `delay` + `backoff` * 2 ** i + Uniform(0, `jitter`))
+    ```
+
+    Where:
+    * i is the retry (poll) number.
+    * Uniform is the uniform distribution.
+
+    Note that the sleep duration is capped by the `max_delay` parameter.
+    """
+
+    backoff: float = 0.9
+    """
+    Exponential backoff factor, in seconds, to use between polls.
+    """
+    delay: float = 0.1
+    """Base delay to use between polls, in seconds."""
     initial_delay: float = 1
-    """Initial delay to use before starting the polling strategy, in
-    seconds."""
+    """
+    Initial delay to use before starting the polling strategy, in seconds.
+    """
     max_delay: float = 20
-    """Maximum delay to use between polls, in seconds. This parameter is
-    activated when the backoff parameter is greater than 1, such that the delay
-    is increasing after each poll."""
+    """Maximum delay to use between polls, in seconds."""
     max_duration: float = 300
     """Maximum duration of the polling strategy, in seconds."""
-    max_tries: int = 20
+    max_tries: int = 100
     """Maximum number of tries to use."""
+    jitter: float = 1
+    """
+    Jitter to use for the polling strategy. A uniform distribution is sampled
+    between 0 and this number. The resulting random number is added to the
+    delay for each poll, adding a random noise. Set this to 0 to avoid using
+    random jitter.
+    """
+    verbose: bool = False
+    """Whether to log the polling strategy. This is useful for debugging."""
 
 
 _DEFAULT_POLLING_OPTIONS: PollingOptions = PollingOptions()
 """Default polling options to use when polling for a run result."""
 
 
-class RunInformation(BaseModel):
-    """Information of a run."""
-
-    description: str
-    """Description of the run."""
-    id: str
-    """ID of the run."""
-    metadata: Metadata
-    """Metadata of the run."""
-    name: str
-    """Name of the run."""
-    user_email: str
-    """Email of the user who submitted the run."""
-
-
-class RunResult(RunInformation):
-    """Result of a run, whether it was successful or not."""
-
-    error_log: Optional[ErrorLog] = None
-    """Error log of the run. Only available if the run failed."""
-    output: Optional[Dict[str, Any]] = None
-    """Output of the run. Only available if the run succeeded."""
-
-
-class RunLog(BaseModel):
-    """Log of a run."""
-
-    log: str
-    """Log of the run."""
-
-
 class UploadURL(BaseModel):
     """Result of getting an upload URL."""
 
@@ -133,13 +103,6 @@ class UploadURL(BaseModel):
     """URL to use for uploading the file."""
 
 
-class Configuration(BaseModel):
-    """Configuration of an instance."""
-
-    execution_class: Optional[str] = None
-    """Execution class for the instance."""
-
-
 @dataclass
 class Application:
     """An application is a published decision model that can be executed."""
@@ -220,13 +183,9 @@ class Application:
             endpoint=f"{self.endpoint}/runs/{run_id}/cancel",
         )
 
-    def delete_batch_experiment(self, batch_id: str) -> None:
+    def delete(self) -> None:
         """
-        Deletes a batch experiment, along with all the associated information,
-        such as its runs.
-
-        Args:
-            batch_id: ID of the batch experiment.
+        Delete the application.
 
         Raises:
             requests.HTTPError: If the response status code is not 2xx.
@@ -234,7 +193,7 @@ class Application:
 
         _ = self.client.request(
             method="DELETE",
-            endpoint=f"{self.experiments_endpoint}/batch/{batch_id}",
+            endpoint=self.endpoint,
         )
 
     def delete_acceptance_test(self, acceptance_test_id: str) -> None:
@@ -254,6 +213,76 @@ class Application:
             endpoint=f"{self.experiments_endpoint}/acceptance/{acceptance_test_id}",
         )
 
+    def delete_batch_experiment(self, batch_id: str) -> None:
+        """
+        Deletes a batch experiment, along with all the associated information,
+        such as its runs.
+
+        Args:
+            batch_id: ID of the batch experiment.
+
+        Raises:
+            requests.HTTPError: If the response status code is not 2xx.
+        """
+
+        _ = self.client.request(
+            method="DELETE",
+            endpoint=f"{self.experiments_endpoint}/batch/{batch_id}",
+        )
+
+    def delete_secrets_collection(self, secrets_collection_id: str) -> None:
+        """
+        Deletes a secrets collection.
+
+        Args:
+            secrets_collection_id: ID of the secrets collection.
+
+        Raises:
+            requests.HTTPError: If the response status code is not 2xx.
+        """
+
+        _ = self.client.request(
+            method="DELETE",
+            endpoint=f"{self.endpoint}/secrets/{secrets_collection_id}",
+        )
+
+    @staticmethod
+    def exists(client: Client, id: str) -> bool:
+        """
+        Check if an application exists.
+
+        Args:
+            client: Client to use for interacting with the Nextmv Cloud API.
+            id: ID of the application.
+
+        Returns:
+            True if the application exists, False otherwise.
+        """
+
+        try:
+            _ = client.request(
+                method="GET",
+                endpoint=f"v1/applications/{id}",
+            )
+            # If the request was successful, the application exists.
+            return True
+        except requests.HTTPError as e:
+            if (
+                # Check whether the error is caused by a 404 status code - meaning the app does not exist.
+                (hasattr(e, "response") and hasattr(e.response, "status_code") and e.response.status_code == 404)
+                or
+                # Check a possibly nested exception as well.
+                (
+                    hasattr(e, "__cause__")
+                    and hasattr(e.__cause__, "response")
+                    and hasattr(e.__cause__.response, "status_code")
+                    and e.__cause__.response.status_code == 404
+                )
+            ):
+                return False
+            # Re-throw the exception if it is not the expected 404 error.
+            raise e from None
+
     def input_set(self, input_set_id: str) -> InputSet:
         """
         Get an input set.
@@ -275,7 +304,28 @@ class Application:
 
         return InputSet.from_dict(response.json())
 
-    def list_acceptance_tests(self) -> List[AcceptanceTest]:
+    def instance(self, instance_id: str) -> Instance:
+        """
+        Get an instance.
+
+        Args:
+            instance_id: ID of the instance.
+
+        Returns:
+            Instance.
+
+        Raises:
+            requests.HTTPError: If the response status code is not 2xx.
+        """
+
+        response = self.client.request(
+            method="GET",
+            endpoint=f"{self.endpoint}/instances/{instance_id}",
+        )
+
+        return Instance.from_dict(response.json())
+
+    def list_acceptance_tests(self) -> list[AcceptanceTest]:
         """
         List all acceptance tests.
 
@@ -293,7 +343,7 @@ class Application:
 
         return [AcceptanceTest.from_dict(acceptance_test) for acceptance_test in response.json()]
 
-    def list_batch_experiments(self) -> List[BatchExperimentMetadata]:
+    def list_batch_experiments(self) -> list[BatchExperimentMetadata]:
         """
         List all batch experiments.
 
@@ -311,7 +361,7 @@ class Application:
 
         return [BatchExperimentMetadata.from_dict(batch_experiment) for batch_experiment in response.json()]
 
-    def list_input_sets(self) -> List[InputSet]:
+    def list_input_sets(self) -> list[InputSet]:
         """
         List all input sets.
 
@@ -329,12 +379,108 @@ class Application:
 
         return [InputSet.from_dict(input_set) for input_set in response.json()]
 
+    def list_instances(self) -> list[Instance]:
+        """
+        List all instances.
+
+        Returns:
+            List of instances.
+
+        Raises:
+            requests.HTTPError: If the response status code is not 2xx.
+        """
+
+        response = self.client.request(
+            method="GET",
+            endpoint=f"{self.endpoint}/instances",
+        )
+
+        return [Instance.from_dict(instance) for instance in response.json()]
+
+    def list_secrets_collections(self) -> list[SecretsCollectionSummary]:
+        """
+        List all secrets collections.
+
+        Returns:
+            List of secrets collections.
+
+        Raises:
+            requests.HTTPError: If the response status code is not 2xx.
+        """
+
+        response = self.client.request(
+            method="GET",
+            endpoint=f"{self.endpoint}/secrets",
+        )
+
+        return [SecretsCollectionSummary.from_dict(secrets) for secrets in response.json()["items"]]
+
+    def list_versions(self) -> list[Version]:
+        """
+        List all versions.
+
+        Returns:
+            List of versions.
+
+        Raises:
+            requests.HTTPError: If the response status code is not 2xx.
+        """
+
+        response = self.client.request(
+            method="GET",
+            endpoint=f"{self.endpoint}/versions",
+        )
+
+        return [Version.from_dict(version) for version in response.json()]
+
+    @classmethod
+    def new(
+        cls,
+        client: Client,
+        name: str,
+        id: Optional[str] = None,
+        description: Optional[str] = None,
+        is_pipeline: Optional[bool] = None,
+    ) -> "Application":
+        """
+        Create a new application.
+
+        Args:
+            client: Client to use for interacting with the Nextmv Cloud API.
+            name: Name of the application.
+            id: ID of the application. Will be generated if not provided.
+            description: Description of the application.
+            is_pipeline: Whether the application is a pipeline.
+
+        Returns:
+            The new application.
+        """
+
+        payload = {
+            "name": name,
+        }
+
+        if description is not None:
+            payload["description"] = description
+        if id is not None:
+            payload["id"] = id
+        if is_pipeline is not None:
+            payload["is_pipeline"] = is_pipeline
+
+        response = client.request(
+            method="POST",
+            endpoint="v1/applications",
+            payload=payload,
+        )
+
+        return cls(client=client, id=response.json()["id"])
+
     def new_acceptance_test(
         self,
         candidate_instance_id: str,
         baseline_instance_id: str,
         id: str,
-        metrics: List[Union[Metric, Dict[str, Any]]],
+        metrics: list[Union[Metric, dict[str, Any]]],
         name: str,
         input_set_id: Optional[str] = None,
         description: Optional[str] = None,
@@ -413,15 +559,77 @@ class Application:
 
         return AcceptanceTest.from_dict(response.json())
 
+    def new_acceptance_test_with_result(
+        self,
+        candidate_instance_id: str,
+        baseline_instance_id: str,
+        id: str,
+        metrics: list[Union[Metric, dict[str, Any]]],
+        name: str,
+        input_set_id: Optional[str] = None,
+        description: Optional[str] = None,
+        polling_options: PollingOptions = _DEFAULT_POLLING_OPTIONS,
+    ) -> AcceptanceTest:
+        """
+        Create a new acceptance test and poll for the result. This is a
+        convenience method that combines the new_acceptance_test with polling
+        logic to check when the acceptance test is done.
+
+        Args:
+            candidate_instance_id: ID of the candidate instance.
+            baseline_instance_id: ID of the baseline instance.
+            id: ID of the acceptance test.
+            metrics: List of metrics to use for the acceptance test.
+            name: Name of the acceptance test.
+            input_set_id: ID of the input set to use for the underlying batch
+                experiment, in case it hasn't been started.
+            description: Description of the acceptance test.
+            polling_options: Options to use when polling for the run result.
+
+        Returns:
+            Result of the acceptance test.
+
+        Raises:
+            requests.HTTPError: If the response status code is not 2xx.
+            TimeoutError: If the acceptance test does not succeed after the
+                polling strategy is exhausted based on time duration.
+            RuntimeError: If the acceptance test does not succeed after the
+                polling strategy is exhausted based on number of tries.
+        """
+        _ = self.new_acceptance_test(
+            candidate_instance_id=candidate_instance_id,
+            baseline_instance_id=baseline_instance_id,
+            id=id,
+            metrics=metrics,
+            name=name,
+            input_set_id=input_set_id,
+            description=description,
+        )
+
+        def polling_func() -> tuple[AcceptanceTest, bool]:
+            test_information = self.acceptance_test(acceptance_test_id=id)
+            if test_information.status in [
+                ExperimentStatus.completed,
+                ExperimentStatus.failed,
+                ExperimentStatus.canceled,
+            ]:
+                return test_information, True
+
+            return None, False
+
+        test_information = poll(polling_options=polling_options, polling_func=polling_func)
+
+        return test_information
+
     def new_batch_experiment(
         self,
         name: str,
         input_set_id: str,
-        instance_ids: List[str] = None,
+        instance_ids: list[str] = None,
         description: Optional[str] = None,
         id: Optional[str] = None,
-        option_sets: Optional[Dict[str, Dict[str, str]]] = None,
-        runs: Optional[List[Union[BatchExperimentRun, Dict[str, Any]]]] = None,
+        option_sets: Optional[dict[str, dict[str, str]]] = None,
+        runs: Optional[list[Union[BatchExperimentRun, dict[str, Any]]]] = None,
     ) -> str:
         """
         Create a new batch experiment.
@@ -475,7 +683,7 @@ class Application:
         end_time: Optional[datetime] = None,
         instance_id: Optional[str] = None,
         maximum_runs: Optional[int] = None,
-        run_ids: Optional[List[str]] = None,
+        run_ids: Optional[list[str]] = None,
         start_time: Optional[datetime] = None,
     ) -> InputSet:
         """
@@ -524,15 +732,63 @@ class Application:
 
         return InputSet.from_dict(response.json())
 
+    def new_instance(
+        self,
+        version_id: str,
+        id: Optional[str] = None,
+        name: Optional[str] = None,
+        description: Optional[str] = None,
+        configuration: Optional[InstanceConfiguration] = None,
+    ) -> Instance:
+        """
+        Create a new instance and associate it with a version.
+
+        Args:
+            version_id: ID of the version to associate the instance with.
+            id: ID of the instance. Will be generated if not provided.
+            name: Name of the instance. Will be generated if not provided.
+            description: Description of the instance. Will be generated if not provided.
+            configuration: Configuration to use for the instance.
+
+        Returns:
+            Instance.
+
+        Raises:
+            requests.HTTPError: If the response status code is not 2xx.
+        """
+
+        payload = {
+            "version_id": version_id,
+        }
+
+        if id is not None:
+            payload["id"] = id
+        if name is not None:
+            payload["name"] = name
+        if description is not None:
+            payload["description"] = description
+        if configuration is not None:
+            payload["configuration"] = configuration.to_dict()
+
+        response = self.client.request(
+            method="POST",
+            endpoint=f"{self.endpoint}/instances",
+            payload=payload,
+        )
+
+        return Instance.from_dict(response.json())
+
     def new_run(  # noqa: C901 # Lot of if statements, but clear logic.
         self,
-        input: Union[Dict[str, Any], BaseModel, str] = None,
+        input: Union[dict[str, Any], BaseModel, str] = None,
         instance_id: Optional[str] = None,
         name: Optional[str] = None,
         description: Optional[str] = None,
         upload_id: Optional[str] = None,
-        options: Optional[Dict[str, str]] = None,
-        configuration: Optional[Configuration] = None,
+        options: Optional[dict[str, str]] = None,
+        configuration: Optional[RunConfiguration] = None,
+        batch_experiment_id: Optional[str] = None,
+        external_result: Optional[ExternalRunResult] = None,
     ) -> str:
         """
         Submit an input to start a new run of the application. Returns the
@@ -548,6 +804,10 @@ class Application:
             upload_id: ID to use when running a large input.
             options: Options to use for the run.
             configuration: Configuration to use for the run.
+            batch_experiment_id: ID of a batch experiment to associate the run
+                with.
+            external_result: External result to use for the run, if this is an
+                external run.
 
         Returns:
             ID of the submitted run.
@@ -561,7 +821,7 @@ class Application:
             input = input.to_dict()
             if input is not None:
                 input_size = get_size(input)
-        elif isinstance(input, Dict):
+        elif isinstance(input, dict):
             input_size = get_size(input)
 
         upload_url_required = isinstance(input, str) or input_size > _MAX_RUN_SIZE
@@ -592,6 +852,10 @@ class Application:
             payload["options"] = options
         if configuration is not None:
             payload["configuration"] = configuration.to_dict()
+        if batch_experiment_id is not None:
+            payload["batch_experiment_id"] = batch_experiment_id
+        if external_result is not None:
+            payload["result"] = external_result.to_dict()
 
         query_params = {
             "instance_id": instance_id if instance_id is not None else self.default_instance_id,
@@ -607,14 +871,16 @@ class Application:
 
     def new_run_with_result(
         self,
-        input: Union[Dict[str, Any], BaseModel] = None,
+        input: Union[dict[str, Any], BaseModel] = None,
         instance_id: Optional[str] = None,
         name: Optional[str] = None,
         description: Optional[str] = None,
         upload_id: Optional[str] = None,
-        run_options: Optional[Dict[str, str]] = None,
+        run_options: Optional[dict[str, str]] = None,
         polling_options: PollingOptions = _DEFAULT_POLLING_OPTIONS,
-        configuration: Optional[Configuration] = None,
+        configuration: Optional[RunConfiguration] = None,
+        batch_experiment_id: Optional[str] = None,
+        external_result: Optional[ExternalRunResult] = None,
     ) -> RunResult:
         """
         Submit an input to start a new run of the application and poll for the
@@ -632,6 +898,10 @@ class Application:
             run_options: Options to use for the run.
             polling_options: Options to use when polling for the run result.
             configuration: Configuration to use for the run.
+            batch_experimemt_id: ID of a batch experiment to associate the run
+                with.
+            external_result: External result to use for the run, if this is an
+                external run
 
          Returns:
             Result of the run.
@@ -652,6 +922,8 @@ class Application:
             upload_id=upload_id,
             options=run_options,
             configuration=configuration,
+            batch_experiment_id=batch_experiment_id,
+            external_result=external_result,
         )
 
         return self.run_result_with_polling(
@@ -659,6 +931,93 @@ class Application:
             polling_options=polling_options,
         )
 
+    def new_secrets_collection(
+        self,
+        secrets: list[Secret],
+        id: Optional[str] = None,
+        name: Optional[str] = None,
+        description: Optional[str] = None,
+    ) -> SecretsCollectionSummary:
+        """
+        Create a new secrets collection. If no secrets are provided, a
+        ValueError is raised.
+
+        Args:
+            secrets: List of secrets to use for the secrets collection. id: ID
+            of the secrets collection. Will be generated if not provided.
+            name: Name of the secrets collection. Will be generated if not
+                provided.
+            description: Description of the secrets collection. Will be
+                generated if not provided.
+
+        Returns:
+            SecretsCollectionSummary: Summary of the secrets collection.
+
+        Raises:
+            ValueError: If no secrets are provided. requests.HTTPError: If the
+            response status code is not 2xx.
+        """
+
+        if len(secrets) == 0:
+            raise ValueError("secrets must be provided")
+
+        payload = {
+            "secrets": [secret.to_dict() for secret in secrets],
+        }
+
+        if id is not None:
+            payload["id"] = id
+        if name is not None:
+            payload["name"] = name
+        if description is not None:
+            payload["description"] = description
+
+        response = self.client.request(
+            method="POST",
+            endpoint=f"{self.endpoint}/secrets",
+            payload=payload,
+        )
+
+        return SecretsCollectionSummary.from_dict(response.json())
+
+    def new_version(
+        self,
+        id: Optional[str] = None,
+        name: Optional[str] = None,
+        description: Optional[str] = None,
+    ) -> Version:
+        """
+        Create a new version using the current dev binary.
+
+        Args:
+            id: ID of the version. Will be generated if not provided.
+            name: Name of the version. Will be generated if not provided.
+            description: Description of the version. Will be generated if not provided.
+
+        Returns:
+            Version.
+
+        Raises:
+            requests.HTTPError: If the response status code is not 2xx.
+        """
+
+        payload = {}
+
+        if id is not None:
+            payload["id"] = id
+        if name is not None:
+            payload["name"] = name
+        if description is not None:
+            payload["description"] = description
+
+        response = self.client.request(
+            method="POST",
+            endpoint=f"{self.endpoint}/versions",
+            payload=payload,
+        )
+
+        return Version.from_dict(response.json())
+
     def push(
         self,
         manifest: Optional[Manifest] = None,
@@ -793,7 +1152,7 @@ class Application:
         except OSError as e:
             raise Exception(f"error deleting output directory: {e}") from e
 
-    def run_input(self, run_id: str) -> Dict[str, Any]:
+    def run_input(self, run_id: str) -> dict[str, Any]:
         """
         Get the input of a run.
 
@@ -831,25 +1190,6 @@ class Application:
 
         return download_response.json()
 
-    def run_logs(self, run_id: str) -> RunLog:
-        """
-        Get the logs of a run.
-
-        Args:
-            run_id: ID of the run.
-
-        Returns:
-            Logs of the run.
-
-        Raises:
-            requests.HTTPError: If the response status code is not 2xx.
-        """
-        response = self.client.request(
-            method="GET",
-            endpoint=f"{self.endpoint}/runs/{run_id}/logs",
-        )
-        return RunLog.from_dict(response.json())
-
     def run_metadata(self, run_id: str) -> RunInformation:
         """
         Get the metadata of a run. The result does not include the run output.
@@ -871,6 +1211,25 @@ class Application:
 
         return RunInformation.from_dict(response.json())
 
+    def run_logs(self, run_id: str) -> RunLog:
+        """
+        Get the logs of a run.
+
+        Args:
+            run_id: ID of the run.
+
+        Returns:
+            Logs of the run.
+
+        Raises:
+            requests.HTTPError: If the response status code is not 2xx.
+        """
+        response = self.client.request(
+            method="GET",
+            endpoint=f"{self.endpoint}/runs/{run_id}/logs",
+        )
+        return RunLog.from_dict(response.json())
+
     def run_result(self, run_id: str) -> RunResult:
         """
         Get the result of a run. The result includes the run output.
@@ -910,38 +1269,108 @@ class Application:
             requests.HTTPError: If the response status code is not 2xx.
         """
 
-        time.sleep(polling_options.initial_delay)
-        delay = polling_options.delay
-        polling_ok = False
-        for _ in range(polling_options.max_tries):
+        def polling_func() -> tuple[any, bool]:
             run_information = self.run_metadata(run_id=run_id)
-            if run_information.metadata.status_v2 in [
+            if run_information.metadata.status_v2 in {
                 StatusV2.succeeded,
                 StatusV2.failed,
                 StatusV2.canceled,
-            ]:
-                polling_ok = True
-                break
+            }:
+                return run_information, True
 
-            if delay > polling_options.max_duration:
-                raise TimeoutError(
-                    f"run {run_id} did not succeed after {delay} seconds",
-                )
+            return None, False
 
-            sleep_duration = min(delay, polling_options.max_delay)
-            time.sleep(sleep_duration)
-            delay *= polling_options.backoff
-
-        if not polling_ok:
-            raise RuntimeError(
-                f"run {run_id} did not succeed after {polling_options.max_tries} tries",
-            )
+        run_information = poll(polling_options=polling_options, polling_func=polling_func)
 
         return self.__run_result(run_id=run_id, run_information=run_information)
 
+    def update_instance(
+        self,
+        id: str,
+        name: str,
+        version_id: Optional[str] = None,
+        description: Optional[str] = None,
+        configuration: Optional[InstanceConfiguration] = None,
+    ) -> Instance:
+        """
+        Update an instance.
+
+        Args:
+            id: ID of the instance to update.
+            version_id: ID of the version to associate the instance with.
+            name: Name of the instance.
+            description: Description of the instance.
+            configuration: Configuration to use for the instance.
+
+        Returns:
+            Instance.
+
+        Raises:
+            requests.HTTPError: If the response status code is not 2xx.
+        """
+
+        payload = {}
+
+        if version_id is not None:
+            payload["version_id"] = version_id
+        if name is not None:
+            payload["name"] = name
+        if description is not None:
+            payload["description"] = description
+        if configuration is not None:
+            payload["configuration"] = configuration.to_dict()
+
+        response = self.client.request(
+            method="PUT",
+            endpoint=f"{self.endpoint}/instances/{id}",
+            payload=payload,
+        )
+
+        return Instance.from_dict(response.json())
+
+    def update_secrets_collection(
+        self,
+        secrets_collection_id: str,
+        name: str,
+        description: str,
+        secrets: list[Secret],
+    ) -> SecretsCollection:
+        """
+        Update a secrets collection.
+
+        Args:
+            secrets_collection_id: ID of the secrets collection.
+            name: Name of the secrets collection.
+            description: Description of the secrets collection.
+            secrets: List of secrets to update.
+
+        Returns:
+            SecretsCollection.
+
+        Raises:
+            ValueError: If no secrets are provided.
+            requests.HTTPError: If the response status code is not 2xx.
+        """
+
+        if len(secrets) == 0:
+            raise ValueError("secrets must be provided")
+
+        payload = {
+            "name": name,
+            "description": description,
+            "secrets": [secret.to_dict() for secret in secrets],
+        }
+        response = self.client.request(
+            method="PUT",
+            endpoint=f"{self.endpoint}/secrets/{secrets_collection_id}",
+            payload=payload,
+        )
+
+        return SecretsCollectionSummary.from_dict(response.json())
+
     def upload_large_input(
         self,
-        input: Union[Dict[str, Any], str],
+        input: Union[dict[str, Any], str],
         upload_url: UploadURL,
     ) -> None:
         """
@@ -955,7 +1384,7 @@ class Application:
             requests.HTTPError: If the response status code is not 2xx.
         """
 
-        if isinstance(input, Dict):
+        if isinstance(input, dict):
             input = json.dumps(input)
 
         _ = self.client.upload_to_presigned_url(
@@ -981,6 +1410,48 @@ class Application:
 
         return UploadURL.from_dict(response.json())
 
+    def secrets_collection(self, secrets_collection_id: str) -> SecretsCollection:
+        """
+        Get a secrets collection.
+
+        Args:
+            secrets_collection_id: ID of the secrets collection.
+
+        Returns:
+            SecretsCollection.
+
+        Raises:
+            requests.HTTPError: If the response status code is not 2xx.
+        """
+
+        response = self.client.request(
+            method="GET",
+            endpoint=f"{self.endpoint}/secrets/{secrets_collection_id}",
+        )
+
+        return SecretsCollection.from_dict(response.json())
+
+    def version(self, version_id: str) -> Version:
+        """
+        Get a version.
+
+        Args:
+            version_id: ID of the version.
+
+        Returns:
+            Version.
+
+        Raises:
+            requests.HTTPError: If the response status code is not 2xx.
+        """
+
+        response = self.client.request(
+            method="GET",
+            endpoint=f"{self.endpoint}/versions/{version_id}",
+        )
+
+        return Version.from_dict(response.json())
+
     def __run_result(
         self,
         run_id: str,
@@ -1049,7 +1520,7 @@ class Application:
                 method="PUT",
                 endpoint=upload_url,
                 data=f,
-                headers={"Content-Type": "application/gzip"},
+                headers={"Content-Type": "application/octet-stream"},
             )
 
         activation_request = {
@@ -1076,3 +1547,74 @@ class Application:
                     indent=2,
                 )
             )
+
+
+def poll(polling_options: PollingOptions, polling_func: callable) -> any:
+    """
+    Auxiliary function for polling.
+
+    The `polling_func` is a callable that must return a `tuple[any, bool]`
+    where the first element is the result of the polling and the second
+    element is a boolean indicating if the polling was successful or should be
+    retried.
+
+    This function will return the result of the `polling_func` if the polling
+    process is successful, otherwise it will raise a `TimeoutError` or
+    `RuntimeError` depending on the situation.
+
+    Parameters
+    ----------
+    polling_options : PollingOptions
+        Options for the polling process.
+    polling_func : callable
+        Function to call to check if the polling was successful.
+
+    Returns
+    -------
+    any
+        Result of the polling function.
+    """
+
+    # Start by sleeping for the duration specified as initial delay.
+    if polling_options.verbose:
+        log(f"polling | sleeping for initial delay: {polling_options.initial_delay}")
+
+    time.sleep(polling_options.initial_delay)
+
+    start_time = time.time()
+
+    # Begin the polling process.
+    for ix in range(polling_options.max_tries):
+        # We check if we can stop polling.
+        result, ok = polling_func()
+        if polling_options.verbose:
+            log(f"polling | try # {ix + 1}, ok: {ok}")
+
+        if ok:
+            return result
+
+        # An exit condition happens if we exceed the allowed duration.
+        passed = time.time() - start_time
+        if polling_options.verbose:
+            log(f"polling | elapsed time: {passed}")
+
+        if passed >= polling_options.max_duration:
+            raise TimeoutError(
+                f"polling did not succeed after {passed} seconds, exceeds max duration: {polling_options.max_duration}",
+            )
+
+        # Calculate the delay.
+        delay = polling_options.delay  # Base
+        delay += polling_options.backoff * (2**ix)  # Add exponential backoff.
+        delay += random.uniform(0, polling_options.jitter)  # Add jitter.
+
+        # Sleep for the calculated delay. We cannot exceed the max delay.
+        sleep_duration = min(delay, polling_options.max_delay)
+        if polling_options.verbose:
+            log(f"polling | sleeping for duration: {sleep_duration}")
+
+        time.sleep(sleep_duration)
+
+    raise RuntimeError(
+        f"polling did not succeed after {polling_options.max_tries} tries",
+    )
