@@ -1,7 +1,6 @@
 """This module contains definitions for input sets."""
 
 from datetime import datetime
-from typing import List
 
 from nextmv.base_model import BaseModel
 
@@ -17,7 +16,7 @@ class InputSet(BaseModel):
     """Description of the input set."""
     id: str
     """ID of the input set."""
-    input_ids: List[str]
+    input_ids: list[str]
     """IDs of the inputs in the input set."""
     name: str
     """Name of the input set."""
