@@ -8,9 +8,9 @@ import os
 import sys
 from dataclasses import dataclass
 from enum import Enum
-from typing import Any, Dict, List, Optional, Union
+from typing import Any, Optional, Union
 
-from pydantic import Field
+from pydantic import AliasChoices, Field
 
 from nextmv.base_model import BaseModel
 from nextmv.logger import reset_stdout
@@ -27,8 +27,8 @@ class RunStatistics(BaseModel):
         Duration of the run in seconds.
     iterations : int, optional
         Number of iterations.
-    custom : Union[Any, Dict[str, Any]], optional
-        Custom statistics created by the user. Can normally expect a `Dict[str,
+    custom : Union[Any, dict[str, Any]], optional
+        Custom statistics created by the user. Can normally expect a `dict[str,
         Any]`.
     """
 
@@ -39,10 +39,10 @@ class RunStatistics(BaseModel):
     custom: Optional[
         Union[
             Any,
-            Dict[str, Any],
+            dict[str, Any],
         ]
     ] = None
-    """Custom statistics created by the user. Can normally expect a `Dict[str,
+    """Custom statistics created by the user. Can normally expect a `dict[str,
     Any]`."""
 
 
@@ -56,8 +56,8 @@ class ResultStatistics(BaseModel):
         Duration of the run in seconds.
     value : float, optional
         Value of the result.
-    custom : Union[Any, Dict[str, Any]], optional
-        Custom statistics created by the user. Can normally expect a `Dict[str,
+    custom : Union[Any, dict[str, Any]], optional
+        Custom statistics created by the user. Can normally expect a `dict[str,
         Any]`.
     """
 
@@ -68,10 +68,10 @@ class ResultStatistics(BaseModel):
     custom: Optional[
         Union[
             Any,
-            Dict[str, Any],
+            dict[str, Any],
         ]
     ] = None
-    """Custom statistics created by the user. Can normally expect a `Dict[str,
+    """Custom statistics created by the user. Can normally expect a `dict[str,
     Any]`."""
 
 
@@ -101,13 +101,13 @@ class Series(BaseModel):
     ----------
     name : str, optional
         Name of the series.
-    data_points : List[DataPoint], optional
+    data_points : list[DataPoint], optional
         Data of the series.
     """
 
     name: Optional[str] = None
     """Name of the series."""
-    data_points: Optional[List[DataPoint]] = None
+    data_points: Optional[list[DataPoint]] = None
     """Data of the series."""
 
 
@@ -119,13 +119,13 @@ class SeriesData(BaseModel):
     ----------
     value : Series, optional
         A series for the value of the solution.
-    custom : List[Series], optional
+    custom : list[Series], optional
         A list of series for custom statistics.
     """
 
     value: Optional[Series] = None
     """A series for the value of the solution."""
-    custom: Optional[List[Series]] = None
+    custom: Optional[list[Series]] = None
     """A list of series for custom statistics."""
 
 
@@ -151,7 +151,11 @@ class Statistics(BaseModel):
     """Statistics about the last result."""
     series_data: Optional[SeriesData] = None
     """Data of the series."""
-    statistics_schema: Optional[str] = Field(alias="schema", default="v1")
+    statistics_schema: Optional[str] = Field(
+        serialization_alias="schema",
+        validation_alias=AliasChoices("schema", "statistics_schema"),
+        default="v1",
+    )
     """Schema (version). This class only supports `v1`."""
 
 
@@ -164,6 +168,73 @@ class OutputFormat(str, Enum):
     """CSV archive format: multiple CSV files."""
 
 
+class VisualSchema(str, Enum):
+    """Schema of a visual asset."""
+
+    CHARTJS = "chartjs"
+    """Tells Nextmv Console to render the custom asset data with the Chart.js
+    library."""
+    GEOJSON = "geojson"
+    """Tells Nextmv Console to render the custom asset data as GeoJSON on a
+    map."""
+    PLOTLY = "plotly"
+    """Tells Nextmv Console to render the custom asset data with the Plotly
+    library."""
+
+
+class Visual(BaseModel):
+    """
+    Visual schema of an asset that defines how it is plotted in the Nextmv
+    Console.
+    """
+
+    visual_schema: VisualSchema = Field(
+        serialization_alias="schema",
+        validation_alias=AliasChoices("schema", "visual_schema"),
+    )
+    """Schema of the visual asset."""
+    label: str
+    """Label for the custom tab of the visual asset in the Nextmv Console."""
+
+    visual_type: Optional[str] = Field(
+        serialization_alias="type",
+        validation_alias=AliasChoices("type", "visual_type"),
+        default="custom-tab",
+    )
+    """Defines the type of custom visual, currently there is only one type:
+    `custom-tab`. This renders the visual in its own tab view of the run
+    details."""
+
+    def __post_init__(self):
+        if self.visual_schema not in VisualSchema:
+            raise ValueError(f"unsupported schema: {self.visual_schema}, supported schemas are {VisualSchema}")
+
+        if self.visual_type != "custom-tab":
+            raise ValueError(f"unsupported visual_type: {self.visual_type}, supported types are `custom-tab`")
+
+
+class Asset(BaseModel):
+    """
+    An asset represents downloadable information that is part of the `Output`.
+    """
+
+    name: str
+    """Name of the asset."""
+    content: Any
+    """Content of the asset. The type must be serializable to JSON."""
+
+    content_type: Optional[str] = "json"
+    """Content type of the asset. Only `json` is allowed"""
+    description: Optional[str] = None
+    """Description of the asset."""
+    visual: Optional[Visual] = None
+    """Visual schema of the asset."""
+
+    def __post_init__(self):
+        if self.content_type != "json":
+            raise ValueError(f"unsupported content_type: {self.content_type}, supported types are `json`")
+
+
 @dataclass
 class Output:
     """
@@ -180,8 +251,8 @@ class Output:
     result of the decision problem. The solution’s type must match the
     `output_format`:
 
-    - `OutputFormat.JSON`: the data must be `Dict[str, Any]`.
-    - `OutputFormat.CSV_ARCHIVE`: the data must be `Dict[str, List[Dict[str,
+    - `OutputFormat.JSON`: the data must be `dict[str, Any]`.
+    - `OutputFormat.CSV_ARCHIVE`: the data must be `dict[str, list[dict[str,
       Any]]]`. The keys represent the file names where the data should be
       written. The values are lists of dictionaries, where each dictionary
       represents a row in the CSV file.
@@ -197,9 +268,9 @@ class Output:
         Options that the `Input` were created with.
     output_format : OutputFormat, optional
         Format of the output data. Default is `OutputFormat.JSON`.
-    solution : Union[Dict[str, Any], Dict[str, List[Dict[str, Any]]], optional
+    solution : Union[dict[str, Any], dict[str, list[dict[str, Any]]], optional
         The solution to the decision problem.
-    statistics : Union[Statistics, Dict[str, Any], optional
+    statistics : Union[Statistics, dict[str, Any], optional
         Statistics of the solution.
     """
 
@@ -209,17 +280,19 @@ class Output:
     """Format of the output data. Default is `OutputFormat.JSON`."""
     solution: Optional[
         Union[
-            Union[Dict[str, Any], Any],  # JSON
-            Dict[str, List[Dict[str, Any]]],  # CSV_ARCHIVE
+            Union[dict[str, Any], Any],  # JSON
+            dict[str, list[dict[str, Any]]],  # CSV_ARCHIVE
         ]
     ] = None
     """The solution to the decision problem."""
-    statistics: Optional[Union[Statistics, Dict[str, Any]]] = None
+    statistics: Optional[Union[Statistics, dict[str, Any]]] = None
     """Statistics of the solution."""
-    csv_configurations: Optional[Dict[str, Any]] = None
+    csv_configurations: Optional[dict[str, Any]] = None
     """Optional configuration for writing CSV files, to be used when the
     `output_format` is OutputFormat.CSV_ARCHIVE. These configurations are
     passed as kwargs to the `DictWriter` class from the `csv` module."""
+    assets: Optional[list[Asset]] = None
+    """Optional list of assets to be included in the output."""
 
     def __post_init__(self):
         """Check that the solution matches the format given to initialize the
@@ -268,26 +341,25 @@ class LocalOutputWriter(OutputWriter):
     """
 
     def _write_json(
-        output: Union[Output, Dict[str, Any]],
-        options: Dict[str, Any],
-        statistics: Dict[str, Any],
+        output: Union[Output, dict[str, Any]],
+        options: dict[str, Any],
+        statistics: dict[str, Any],
+        assets: list[dict[str, Any]],
         path: Optional[str] = None,
     ) -> None:
-        solution = {}
-        if isinstance(output, Output):
-            sol = output.solution
-        elif isinstance(output, Dict):
-            sol = output.get("solution")
-
-        if sol is not None:
-            solution = sol
-
-        serialized = json.dumps(
-            {
+        if isinstance(output, dict):
+            final_output = output
+        else:
+            solution = output.solution if output.solution is not None else {}
+            final_output = {
                 "options": options,
                 "solution": solution,
                 "statistics": statistics,
-            },
+                "assets": assets,
+            }
+
+        serialized = json.dumps(
+            final_output,
             indent=2,
             default=_custom_serial,
         )
@@ -301,8 +373,9 @@ class LocalOutputWriter(OutputWriter):
 
     def _write_archive(
         output: Output,
-        options: Dict[str, Any],
-        statistics: Dict[str, Any],
+        options: dict[str, Any],
+        statistics: dict[str, Any],
+        assets: list[dict[str, Any]],
         path: Optional[str] = None,
     ) -> None:
         dir_path = "output"
@@ -319,6 +392,7 @@ class LocalOutputWriter(OutputWriter):
             {
                 "options": options,
                 "statistics": statistics,
+                "assets": assets,
             },
             indent=2,
         )
@@ -350,7 +424,7 @@ class LocalOutputWriter(OutputWriter):
 
     def write(
         self,
-        output: Union[Output, Dict[str, Any]],
+        output: Union[Output, dict[str, Any]],
         path: Optional[str] = None,
         skip_stdout_reset: bool = False,
     ) -> None:
@@ -370,9 +444,14 @@ class LocalOutputWriter(OutputWriter):
         unexpected behavior. If you want to skip this behavior, set the
         `skip_stdout_reset` parameter to `True`.
 
+        If the `output` is a `dict`, it will be simply written to the specified
+        `path`, as a passthrough. On the other hand, if the `output` is of type
+        `Output`, a more structured object will be written, which adheres to
+        the schema specified by the corresponding `Output` class.
+
         Parameters
         ----------
-        output: Output, Dict[str, Any]
+        output: Output, dict[str, Any]
             Output data to write.
         path : str
             Path to write the output data to.
@@ -393,70 +472,96 @@ class LocalOutputWriter(OutputWriter):
 
         if isinstance(output, Output):
             output_format = output.output_format
-        elif isinstance(output, Dict):
+        elif isinstance(output, dict):
             output_format = OutputFormat.JSON
         else:
-            raise TypeError(f"unsupported output type: {type(output)}, supported types are `Output` or `Dict`")
+            raise TypeError(f"unsupported output type: {type(output)}, supported types are `Output` or `dict`")
 
         statistics = self._extract_statistics(output)
         options = self._extract_options(output)
+        assets = self._extract_assets(output)
 
         self.FILE_WRITERS[output_format](
             output=output,
             options=options,
             statistics=statistics,
+            assets=assets,
             path=path,
         )
 
     @staticmethod
-    def _extract_statistics(output: Union[Output, Dict[str, Any]]) -> Dict[str, Any]:
+    def _extract_statistics(output: Union[Output, dict[str, Any]]) -> dict[str, Any]:
         """Extract JSON-serializable statistics."""
 
         statistics = {}
 
-        if isinstance(output, Output):
-            stats = output.statistics
-        elif isinstance(output, Dict):
-            stats = output.get("statistics")
+        if not isinstance(output, Output):
+            return statistics
+
+        stats = output.statistics
 
         if stats is None:
             return statistics
 
         if isinstance(stats, Statistics):
             statistics = stats.to_dict()
-        elif isinstance(stats, Dict):
+        elif isinstance(stats, dict):
             statistics = stats
         else:
-            raise TypeError(f"unsupported statistics type: {type(stats)}, supported types are `Statistics` or `Dict`")
+            raise TypeError(f"unsupported statistics type: {type(stats)}, supported types are `Statistics` or `dict`")
 
         return statistics
 
     @staticmethod
-    def _extract_options(output: Union[Output, Dict[str, Any]]) -> Dict[str, Any]:
+    def _extract_options(output: Union[Output, dict[str, Any]]) -> dict[str, Any]:
         """Extract JSON-serializable options."""
 
         options = {}
 
-        if isinstance(output, Output):
-            opt = output.options
-        elif isinstance(output, Dict):
-            opt = output.get("options")
+        if not isinstance(output, Output):
+            return options
+
+        opt = output.options
 
         if opt is None:
             return options
 
         if isinstance(opt, Options):
             options = opt.to_dict()
-        elif isinstance(opt, Dict):
+        elif isinstance(opt, dict):
             options = opt
         else:
-            raise TypeError(f"unsupported options type: {type(opt)}, supported types are `Options` or `Dict`")
+            raise TypeError(f"unsupported options type: {type(opt)}, supported types are `Options` or `dict`")
 
         return options
 
+    @staticmethod
+    def _extract_assets(output: Union[Output, dict[str, Any]]) -> list[dict[str, Any]]:
+        """Extract JSON-serializable assets."""
+
+        assets = []
+
+        if not isinstance(output, Output):
+            return assets
+
+        assts = output.assets
+
+        if assts is None:
+            return assets
+
+        for ix, asset in enumerate(assts):
+            if isinstance(asset, Asset):
+                assets.append(asset.to_dict())
+            elif isinstance(asset, dict):
+                assets.append(asset)
+            else:
+                raise TypeError(f"unsupported asset {ix}, type: {type(asset)}; supported types are `Asset` or `dict`")
+
+        return assets
+
 
 def write_local(
-    output: Union[Output, Dict[str, Any]],
+    output: Union[Output, dict[str, Any]],
     path: Optional[str] = None,
     skip_stdout_reset: bool = False,
 ) -> None:
@@ -481,7 +586,7 @@ def write_local(
 
     Parameters
     ----------
-    output : Output, Dict[str, Any]
+    output : Output, dict[str, Any]
         Output data to write.
     path : str
         Path to write the output data to.
